# Copyright 2018 Eggplant
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for the Eggplant AI Python client library

See `<http://docs.testplant.com/EAI/eai-rest-api.htm>`
for full details of the API calls and messages.
"""

import json

import requests

class AIClientException(Exception):
    """
    Class for exceptions raised by the AI API client
    """
    pass

class AIClient(object):
    """
    Client for the Eggplant AI REST API
    """
    def __init__(self, url, username, password):
        """
        Construct with a URL, username, and password

        :param url: The base URL, e.g., `<http://localhost>`_
        :param username:
        :param password:
        """
        self.url = url
        self.username = username
        self.password = password
        self.access_token = None
        self._get_access_token()


    def _do_post(self, endpoint, payload, operation):
        """
        Perform an authenticated HTTP POST request, returning the JSON result

        :param endpoint: The endpoint path, e.g., '/api/run'
        :param payload: The urlencoded request body
        :param operation: Text description of the operation, for error reporting
        :returns: The JSON response
        """
        headers = {'Authorization' : 'bearer ' + self.access_token}
        auth_failed = False
        while True:
            try:
                r = requests.post(self.url + endpoint, headers=headers, data=payload)
            except requests.exceptions.ConnectionError as er:
                raise AIClientException("Failed to {}: {}".format(operation, er))
            if r.status_code == 401:
                # Invalid auth token
                if auth_failed:
                    # Don't try again if the new auth token doesn't work
                    raise AIClientException("Authentication failed: {}".format(r.text))
                else:
                    # Auth token has expired: get a new one
                    auth_failed = True
                    self._get_access_token()
            elif r.status_code != 200:
                # Any other error
                raise AIClientException("Failed to {}: {}".format(operation, r.text))
            else:
                # Success
                return r.json()


    def _get_access_token(self):
        """
        Get an API access token
        """
        payload = {'username' : self.username, 'password' : self.password}
        auth_url = '/auth'
        error_message = 'Failed to get access token: {}'
        try:
            r = requests.post(self.url + auth_url, data=payload)
        except requests.exceptions.ConnectionError as er:
            raise AIClientException(error_message.format(er))
        if r.status_code != 200:
            raise AIClientException(error_message.format(r.text))
        response = r.json()
        self.access_token = response['access_token']


    def run_model(self, model, group, directed_test=None, agent_name=None , suite_location=None, bins=100,
            coverdb='ttdb', covertarget=1, execute=True, iterations=1, logdir='logs', logfile=None,
            max_actions=10000, onerror='clean', path=None, replay=None, seed=None, verbosity='0x300'):
        """
        Run a model

        :param model:
        :param group:
        :param directed_test: Name of a directed test to run. Default: None
        :param agent_name: The Eggplant AI Agent name. Default: None
        :param suite_location: The Eggplant Functinal suite location. Default: None
        :param bins: Number of bins. Default: 100
        :param coverdb: The name of the coverage database, Default: 'ttdb'
        :covertarget: The target coverage order. Default: 1
        :param execute: Execute SenseTalk snippets? Default: True
        :param iterations: Number of iterations: Default: 1
        :param logdir: Log file directory. Default: 'logs'
        :param logfile: Log filename. Default: None (uses 'modelname_seed_date.log')
        :param max_actions: Maximum number of actions. Default: 10000
        :param onerror: Action to take on error. One of 'clean', 'continue', 'immediate', or 'debug'. Default: 'clean'
        :param path: Path. Default: None
        :param replay: Test ID to replay. Default: None
        :param seed: Test seed. Default: None
        :param verbosity: Verbosity. Default: '0x300'
        :returns: The JSON response
        """
        payload = {'modelName': model, 'group': group}
        if directed_test is not None:
            payload['directedTest'] = directed_test
        if agent_name is not None:
            payload['agentName'] = agent_name
        if suite_location is not None:
            payload['suiteLocation'] = suite_location
        if bins is not None and bins != 100:
            payload['bins'] = bins
        if coverdb is not None and coverdb != 'ttdb':
            payload['coverdb'] = coverdb
        if covertarget is not None and covertarget != 1:
            payload['covertarget'] = covertarget
        if execute is not None and execute != True:
            payload['execute'] = execute
        if iterations is not None and iterations != 1:
            payload['iterations'] = iterations
        if logdir is not None and logdir != 'logs':
            payload['logdir'] = logdir
        if logfile is not None:
            payload['logfile'] = logfile
        if max_actions is not None and max_actions != 1000:
            payload['maxActions'] = max_actions
        if onerror is not None and onerror != 'clean':
            payload['onerror'] = onerror
        if path is not None:
            payload['path'] = path
        if replay is not None:
            payload['replay'] = replay
        if seed is not None:
            payload['seed'] = seed
        if verbosity is not None and verbosity != '0x300':
            payload['verbosity'] = verbosity
        return self._do_post('/api/run', payload, "run model")


    def get_progress(self, id):
        """
        Get the progress of a run

        :param id: The run id as returned in the `id` field by `runModel()`
        :returns: The JSON response
        """
        payload = {'id': id}
        return self._do_post('/api/run/progress', payload, "get progress")


    def get_sut_status(self):
        """
        Get the status of the SUT

        :returns: The JSON response
        """
        return self._do_post('/api/sut/status', None, "get SUT status")


    def abort_run(self, process_id):
        """
        Abort a run

        :param process_id: The process id, as returned in the `processid` field by `runModel()`
        :returns: The JSON response
        """
        payload = {'processid': process_id}
        return self._do_post('/api/run/abort', payload, "abort run")


    def get_run_report(self, model, group):
        """
        Get the run report

        :param model:
        :param group:
        :returns The response JSON
        """
        payload = {'modelName' : model, 'group': group}
        return self._do_post('/api/run/report', payload, "get run report")


    def get_coverage_report(self, model, group):
        """
        Get the coverage report

        :param model:
        :param group:
        :returns: The response JSON
        """
        payload = {'modelName' : model, 'group': group}
        return self._do_post('/api/coverage/report', payload, "get coverage report")


    def get_testcase_report(self, model, group, testcases):
        """
        Get the testcase report

        :param model:
        :param group:
        :param testcases: A list of testcases
        :returns: The response JSON
        """
        payload = {'modelName': model, 'group': group, 'testcases': testcases}
        return self._do_post('/api/testcase/report', payload, "get testcase report")

